/*
 * modbus_tiny.h
 *
 *      Author: arkhnchul
 */

#ifndef INC_MODBUS_TINY_H_
#define INC_MODBUS_TINY_H_

#include <stdint.h>

/* If target MCU is big endian, it's possible to access two-byte fields of ADU
 * in more straightforward way */
#define MB_MCU_BIG_ENDIAN

/* Size of communication buffer in bytes. No ADU larger than this can be received.
 * Modbus spec declares maximum ADU size of 256 bytes, but on modest MCU
 * we can violate it and save precious RAM */
#define MB_BUFFER_SIZE 64

/*
 * Modbus functions supported by slave
 * */
//#define MB_HAS_FUNCTION_DIAGNOSTICS
#define MB_HAS_FUNCTION_READ_HOLDING_REGISTERS
/* Read Input Registers is effectively obsolete. Here it provides read acces to the same holding registers */
//#define MB_HAS_FUNCTION_READ_INPUT_REGISTERS
//#define MB_HAS_FUNCTION_READ_COILS
/* Here it provides read acces to coils */
//#define MB_HAS_FUNCTION_READ_DISCRETE_INPUTS
/* all registers are read-only in default implementation, see mb_port_write_register() source */
//#define MB_HAS_FUNCTION_WRITE_SINGLE_REGISTER
/* all coils are read-only in default implementation, see mb_port_write_coil() source */
//#define MB_HAS_FUNCTION_WRITE_SINGLE_COIL

/* Possible slave status */
typedef enum {
    MB_STATE_IDLE, /* nothing interesting is going on */
    MB_STATE_RECEIVING, /* slave already received first byte with it's address and continues to put ADU bytes in buffer */
    MB_STATE_TRANSMITTING, /* slave transmits response to Master */
    MB_STATE_PROCESSING, /* slave tries to process request */
    MB_STATE_IGNORE, /* somehow bad frame had been received and still transmitting on line */
    MB_STATE_REQUEST_READY, /* whole request ADU had been put in buffer */
    MB_STATE_RESPONSE_READY /* slave had completed request processing and response body ready to be transmitted */
} mb_state_t;

/* modbus exception codes */
typedef enum {
    MB_EXCEPTION_FUNCTION_UNSUPPORTED = (uint8_t) 0x01,
    MB_EXCEPTION_WRONG_ADDRESS = (uint8_t) 0x02,
    MB_EXCEPTION_WRONG_VALUE = (uint8_t) 0x03,
    MB_EXCEPTION_INTERNAL_FAILURE = (uint8_t) 0x04
} mb_exception_t;

/* modbus function codes */
typedef enum {
    MB_FUNC_READ_COILS = (uint8_t) 1,
    MB_FUNC_READ_DISCRETE_INPUTS = (uint8_t) 2,
    MB_FUNC_READ_HOLDING_REGISTERS = (uint8_t) 3,
    MB_FUNC_READ_INPUT_REGISTERS = (uint8_t) 4,
    MB_FUNC_WRITE_SINGLE_COIL = (uint8_t) 5,
    MB_FUNC_WRITE_SINGLE_REGISTER = (uint8_t) 6,
    MB_FUNC_DIAGNOSTICS = (uint8_t) 8
} mb_function_t;

typedef struct {
    uint8_t slave_addr;
    uint8_t ADU_length;
    uint8_t state;
    uint8_t buffer[MB_BUFFER_SIZE];
} mb_data_t;

/*
 * PORT FUNCTIONS
 *
 * mb_port_* functions should be implemented for specific hardware
 *
 */
void mb_port_init();
void mb_port_rx_interframe_timeout();
void mb_port_transmit_response();
uint16_t mb_port_read_register(uint16_t register_addr);
void mb_port_write_register(uint16_t register_addr, uint16_t value);
uint8_t mb_port_read_coil(uint16_t coil_addr);
void mb_port_write_coil(uint16_t register_addr, uint8_t value);

/*
 * SLAVE DATA STRUCTURES
 */
extern volatile mb_data_t MB;

/*
 * PLATFORM INDEPENDENT FUNCTIONS
 */
void mb_init();
void mb_cycle_tick(void);
void mb_rx_byte(uint8_t byte);
void mb_rx_completed(void);
void mb_process_request(void);
void mb_response_ready(void);
void mb_raise_exception(mb_exception_t exception);
uint16_t mb_crc_compute(void);
void mb_crc_fill(void);
void mb_transmit_completed(void);

void mb_function_diagnostic(void);
void mb_function_read_registers(void);
void mb_function_write_single_register(void);
void mb_function_read_coils(void);
void mb_function_write_single_coil(void);

/* various helper macro */
#define MB_ADU_ADDRESS 		MB.buffer[0]
#define MB_ADU_FUNCTION 	MB.buffer[1]
#define MB_ADU_SUBFUNCTION 	MB.buffer[2]
#define MB_ADU_EXCEPTION 	MB.buffer[2]
#define MB_MSB(x) ( (uint8_t) ((x & 0xFF00) >> 8) )
#define MB_LSB(x) ( (uint8_t) ((0xFF & (x))) )
#define MB_COMBINE_UINT16(msb,lsb) ( (((uint16_t)(msb)) << 8) | (lsb) )

#ifdef MB_MCU_BIG_ENDIAN
#define MB_ADU_GET_UINT16(index) ( *((uint16_t*)&(MB.buffer[(index)])) )
#define MB_ADU_SET_UINT16(index,value) ( *((uint16_t*)&(MB.buffer[(index)])) = value )
#else
#define MB_ADU_GET_UINT16(index) ( MB_COMBINE_UINT16(MB.buffer[(index)], MB.buffer[(index)+1]) )
#define MB_ADU_SET_UINT16(index,value) { MB.buffer[(index)]=MB_MSB((value)); MB.buffer[(index)+1]=MB_LSB((value)); }
#endif

#endif /* INC_MODBUS_TINY_H_ */
