/*
 * modbus_tiny_port_stm8.c
 *
 *      Author: arkhnchul
 *
 * Modbus RTU peripheral functions example for STM8 MCU
 * using UART1 as communication device.
 *
 * Modbus spec declares inter-character and inter-frame gaps
 * of less than 1.5 and greater than 3.5 character times (or 1.75ms for
 * baudrates greater than 19200) respectively.
 *
 * Despite the statement "Partial messages must be detected and errors must
 * be set as a result" the very same spec says that incomplete frames or
 * those which have wrong checksum must be silently discarded by slave.
 *
 * Here we have a system timesource of 1ms ticks on TIM4, and use it for
 * checking interframe gaps. Variable "mb_interframe_ttl" decreases with every
 * system tick and set to RX_TTL_RESET_VALUE with every RXE interrupt (==received character).
 * When it comes to zero, we assume that timeout of around 2ms from last received character
 * reached and whole ADU is received.
 *
 * Inter-character gaps is not checked at all, incomplete message will be
 * ignored anyways because of CRC mismatch.
 *
 */

#include <stdint.h>
#include "stm8s.h"
#include "stm8s_uart1.h"
#include "stm8s_itc.h"
#include "modbus_tiny.h"

#define CLOCKFREQ 16000000
#define MB_UART1_BAUDRATE 19200
/* stm8 specific voodoo */
#define UART1DIV (CLOCKFREQ/MB_UART1_BAUDRATE)
#define UART1BRR1 ((UART1DIV & 0x0FF0)>>4)
#define UART1BRR2 (((UART1DIV+1) & 0x000F)|((UART1DIV & 0xF000)>>8))

#define MB_UART1_PARITY UART1_PARITY_EVEN
/* modbus spec declares 11-bit frames, which means 1 stop bit for any parity
 * and 2 stop bits if no parity is used */
#define MB_UART1_STOPBITS UART1_STOPBITS_1
/* ST thinks that "uart word length" means "together with parity bit".
 * Because of pony. */
#define MB_UART1_WORDLENGTH UART1_WORDLENGTH_9D
/* modbus slave address stored in the EEPROM on this address */
#define MB_SLAVE_ADDR_EEPROM FLASH_DATA_START_PHYSICAL_ADDRESS+1

volatile uint8_t mb_interframe_ttl = 0;

void mb_port_init() {
    mb_init();
    /* assorted UART1 config bits */
    UART1->CR1 |= MB_UART1_WORDLENGTH | MB_UART1_PARITY;
    UART1->CR3 |= MB_UART1_STOPBITS;

    /* Set the UART1 BaudRate */
    UART1->BRR2 = UART1BRR2;
    UART1->BRR1 = UART1BRR1;

    /* Set the Transmitter and Receiver Enable bits, RXNE Interrupt Enable bit */
    UART1->CR2 |= (uint8_t) UART1_CR2_TEN | UART1_CR2_REN | UART1_CR2_RIEN;

    /* read Modbus slave address stored in EEPROM */
    MB.slave_addr = *((uint8_t *) MB_SLAVE_ADDR_EEPROM);
}

void mb_port_transmit_response() {
    uint8_t tx_idx = 0;
    while (tx_idx < MB.ADU_length) {
        /* put ADU character into transmission buffer register */
        UART1->DR = MB.buffer[tx_idx];
        /* wait until the end of character transmission */
        while (!(UART1->SR & UART1_FLAG_TXE))
            ;
        tx_idx++;
    }
    mb_transmit_completed();
}

/* not exactly timeout, just ttl tick
 * called by system clock interrupt routine elsewhere */
void mb_port_rx_interframe_timeout() {
    if (mb_interframe_ttl) {
        mb_interframe_ttl--;
        if (!mb_interframe_ttl) {
            mb_rx_completed();
        }
    }
}

#if MB_UART1_BAUDRATE >= 19200
#define RX_TTL_RESET_VALUE 2 //2ms is close enough to demanded 1.75
#else
#define RX_TTL_RESET_VALUE (uint8_t)(((1000 * 39) / MB_UART1_BAUDRATE) + 1)
#endif

INTERRUPT_HANDLER( uart1_rxe_handler, ITC_IRQ_UART1_RX) {
    mb_rx_byte(UART1->DR);
    mb_interframe_ttl = RX_TTL_RESET_VALUE;
}
